/***************************************************************************
                          odbcclientobj.cpp  -  description
                             -------------------
    begin                : Sat Oct 30 1999
    copyright            : (C) 1999 by Team FXML
    email                : fxml@exite.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "stdafx.h"
#include "ODBCClient.h"
#include "ODBCClientObj.h"
#include "ws-util.h"
#include <winsock2.h>

/////////////////////////////////////////////////////////////////////////////
// CODBCClient

///////////////////////////////////////////////////////
//
//	Name: ExecSQL
//	Params: bsSQL - SQL to execute
//			bsResult - result of SQL in XML
//	Returns: S_OK
//	Description: Only exposed method.  Executes SQL
//		and returns result
///////////////////////////////////////////////////////
STDMETHODIMP CODBCClient::ExecSQL(BSTR bsSQL, BSTR *bsResult)
{
	USES_CONVERSION;
	CComBSTR bsRet = L"<?xml version=\"1.0\"?>\r\n";

	LogEvent(((std::string)"ExecSQL reports initializing with SQL: " + (std::string)OLE2T(bsSQL)).c_str(),1);

	//init winsock
    WSAData wsaData;
    if (WSAStartup(MAKEWORD(1, 1), &wsaData) != 0) {
		bsRet +="<result status=\"failure\">\r\n<error>Could not initialize WinSock</error>\r\n</result>\r\n";
		(*bsResult) = SysAllocString(bsRet);
        return S_OK;
    }

	//lookup host address
	u_long nRemoteAddress = LookupAddress(OLE2T(m_bsHostName));
    if (nRemoteAddress == INADDR_NONE) {
        LogEvent(WSAGetLastErrorMessage("lookup address"), 2);
		bsRet +="<result status=\"failure\">\r\n<error>Could not lookup client address</error>\r\n</result>\r\n";
		(*bsResult) = SysAllocString(bsRet);
		return S_OK;
    }
    in_addr Address;
    memcpy(&Address, &nRemoteAddress, sizeof(u_long)); 

	LogEvent(((std::string)"ExecSQL reports attempting to connect to host: " + (std::string)inet_ntoa(Address)).c_str(),1);

	//connect to host
    SOCKET sd = EstablishConnection(nRemoteAddress, htons(m_nPort));
    if (sd == INVALID_SOCKET) {
        LogEvent(WSAGetLastErrorMessage("connect to server"),3);
		bsRet +="<result status=\"failure\">\r\n<error>Could not connect to remote host</error>\r\n</result>\r\n";
		(*bsResult) = SysAllocString(bsRet);
		return S_OK;
    }

	LogEvent(((std::string)"ExecSQL reports connected to host: " + (std::string)inet_ntoa(Address)).c_str(),1);

	//perform SQL query
	bsRet = T2OLE((ReadAndWriteToSocket(sd, bsSQL)).c_str());
	(*bsResult) = SysAllocString(bsRet);

	//shutdown connection
	ShutdownConnection(sd);
	
	// Shut Winsock back down and take off.
    WSACleanup();
	
	return S_OK;
}

///////////////////////////////////////////////////////
//
//	Name: LogEvent
//	Params: pFormat - Message string
//		nSeverity - severity of event
//	Returns: none
//	Description: Logs Event to event log IF severity is >= gnDebugLevel
//nSeverity = 1 for informational
//			= 2 for warnings
//			= 3 for errors
///////////////////////////////////////////////////////
void CODBCClient::LogEvent(LPCTSTR pFormat, short nSeverity)
{
    TCHAR    chMsg[2048];
    HANDLE  hEventSource;
    LPTSTR  lpszStrings[1];
    va_list pArg;

	//ensure event is severe enough to log
	if (nSeverity < m_nDebugLevel) return;
	if (strlen(pFormat) > 2040) return;

    va_start(pArg, pFormat);
    _vstprintf(chMsg, pFormat, pArg);
    va_end(pArg);

    lpszStrings[0] = chMsg;

     /* Get a handle to use with ReportEvent(). */
    hEventSource = RegisterEventSource(NULL, "ODBCClient");
    if (hEventSource != NULL)
    {
        /* Write to event log. */
		WORD wEventSource = EVENTLOG_INFORMATION_TYPE;
		if (nSeverity == 1) wEventSource = EVENTLOG_INFORMATION_TYPE;
		else if (nSeverity == 2) wEventSource = EVENTLOG_WARNING_TYPE;
		else if (nSeverity == 3) wEventSource = EVENTLOG_ERROR_TYPE;

        ReportEvent(hEventSource, wEventSource, 0, 0, NULL, 1, 0, (LPCTSTR*) &lpszStrings[0], NULL);
        DeregisterEventSource(hEventSource);
    }

 }

//////////////////////////////////////////////////////
//
//	Name: EstablishConnection
//	Params: nRemoteAddr - remote address to connect to 
//			nPort - port to connect to
//	Returns: socket result
//	Description: Connects to a given address, on a given port, both of which must be
// in network byte order.  Returns newly-connected socket if we succeed,
// or INVALID_SOCKET if we fail.
///////////////////////////////////////////////////////
SOCKET CODBCClient::EstablishConnection(u_long nRemoteAddr, u_short nPort)
{
    // Create a stream socket
    SOCKET sd = socket(AF_INET, SOCK_STREAM, 0);
    if (sd != INVALID_SOCKET) {
        sockaddr_in sinRemote;
        sinRemote.sin_family = AF_INET;
        sinRemote.sin_addr.s_addr = nRemoteAddr;
        sinRemote.sin_port = nPort;
        if (connect(sd, (sockaddr*)&sinRemote, sizeof(sockaddr_in)) ==
                SOCKET_ERROR) {
            sd = INVALID_SOCKET;
        }
    }

    return sd;
}

//////////////////////////////////////////////////////
//
//	Name: ReadAndWriteToSocket
//	Params: sd - socket to write/read to 
//			bsSQL - SQL to execute 
//	Returns: string - XML SQL result
//	Description: Writes the XML string to the socket, reads
//		back the sql result and returns it
///////////////////////////////////////////////////////

std::string CODBCClient::ReadAndWriteToSocket(SOCKET sd, BSTR bsSQL)
{
    // Send the string to the server
	//variables for setting up timeout
	USES_CONVERSION;

	fd_set read_fds; 
	struct timeval TimeOut; 
	FD_ZERO(&read_fds); 
	FD_SET(sd, &read_fds); 
	TimeOut.tv_sec = m_nSocketTimeout; //use timeout of durartion m_nSocketTimeout
	TimeOut.tv_usec = 0; 
	std::string sXML; //xml to send
	std::string sReturn, sTemp; //xml to receive
	int nReadBytes; //number of bytes returned
	BOOL bAllDone = FALSE; //socket read var
	short nFind; //find var

	//init return val
	sReturn = "";

	//now, replace all the &, > and < in the originial SQL because we want
	//to send valid XML
	sTemp = OLE2T(bsSQL);

	//sanity check
	if (sTemp.length() == 0) return sReturn;

	sTemp = SubStrReplace(sTemp, "&", "&amp;");
	sTemp = SubStrReplace(sTemp, "<", "&lt;");
	sTemp = SubStrReplace(sTemp, ">", "&gt;");


	//first, construct XML string to send
	sXML = "<?xml version=\"1.0\"?>\r\n<request>\r\n<connectionstring>" + 
		(std::string)OLE2T(m_bsConnectionString) + (std::string)"</connectionstring>\r\n<sql>" +
		sTemp + "</sql>\r\n</request>\r\n";

	const int kEchoMessageLen = sXML.length();
	if (send(sd, sXML.c_str(), kEchoMessageLen, 0) == SOCKET_ERROR) {
		LogEvent("ReadAndWriteToSocket reports socket write error", 2);
		return "";
	}

			// Read reply from server
	do
	{
		if (select(FD_SETSIZE, &read_fds, NULL, NULL, &TimeOut) != SOCKET_ERROR) 
		{ 
			if (FD_ISSET(sd, &read_fds)) 
			{ 
				char *cBuf = (char*)malloc(1000);
				nReadBytes = recv(sd, cBuf, 1000, 0);
				if ( nReadBytes > 0)
				{
					sTemp = cBuf;
					sReturn += sTemp.substr(0, nReadBytes);
					free(cBuf);
					cBuf = (char*)malloc(1000);
				} //if nreadbytes > 0
				//look for end of request
				nFind = sReturn.find("</result>");
				if (nFind > 0) bAllDone = TRUE;
			}//if FD_ISSET
			else
			{
				sReturn = "<?xml version=\"1.0\"?>\r\n<result status=\"failure\">\r\n<error>Connection timed out waiting for response</error>\r\n</result>";
				nReadBytes = -1;
			}
		}//select
		else
		{
			sReturn = "<?xml version=\"1.0\"?>\r\n<result status=\"failure\">\r\n<error>Connection timed out waiting for response</error>\r\n</result>";
			nReadBytes = -1;
		}
	} while ((nReadBytes > 0) && (!bAllDone));
 
	if (sReturn.empty()) sReturn = "<?xml version=\"1.0\"?>\r\n<result status=\"failure\">\r\n<error>Communication error, connection possibly denied</error>\r\n</result>";
    return sReturn;
}




//////////////////////////////////////////////////////
//
//	Name: LookupAddress
//	Params: pcHost - address to lookup 
//	Returns: ip address
//	Description: Given an address string, determine if it's a dotted-quad IP address
// or a domain address.  If the latter, ask DNS to resolve it.  In
// either case, return resolved IP address.  If we fail, we return
// INADDR_NONE.
///////////////////////////////////////////////////////

u_long CODBCClient::LookupAddress(const char* pcHost)
{
    u_long nRemoteAddr = inet_addr(pcHost);
    if (nRemoteAddr == INADDR_NONE) {
        // pcHost isn't a dotted IP, so resolve it through DNS
        hostent* pHE = gethostbyname(pcHost);
        if (pHE == 0) {
            return INADDR_NONE;
        }
        nRemoteAddr = *((u_long*)pHE->h_addr_list[0]);
    }

    return nRemoteAddr;
}



///////////////////////////////////////////////////////
//
//	Name: SubStrReplace
//	Params: cOriginal - original string to search
//			cSearch - char to search for
//			cReplace - char to replace
//	Returns: string with cSearch replaced by cReplace
//	Description: Searches and replaces occurences of cSearch
//		in cOriginal with cReplace
///////////////////////////////////////////////////////
std::string CODBCClient::SubStrReplace(std::string cOriginal, char* cSearch, char* cReplace)
{
	std::string sRet = cOriginal; //set return string
	int nFind; //location of found string

	//search for char string
	nFind = sRet.find(cSearch, 0);
	while (nFind >= 0)
	{
		sRet = sRet.substr(0, nFind) + (std::string)cReplace + sRet.substr(nFind + strlen(cSearch), sRet.length() - nFind);

		nFind = sRet.find(cSearch, nFind + strlen(cReplace));
	}

	return sRet;
}

//data wrappers.  Boring :)

STDMETHODIMP CODBCClient::get_HostName(BSTR *pVal)
{
	(*pVal) = SysAllocString(m_bsHostName);

	return S_OK;
}

STDMETHODIMP CODBCClient::put_HostName(BSTR newVal)
{
	if (m_bsHostName) SysFreeString(m_bsHostName);
	m_bsHostName = SysAllocString(newVal);

	return S_OK;
}

STDMETHODIMP CODBCClient::get_Port(short *pVal)
{
	(*pVal) = m_nPort;

	return S_OK;
}

STDMETHODIMP CODBCClient::put_Port(short newVal)
{
	m_nPort = newVal;

	return S_OK;
}

STDMETHODIMP CODBCClient::get_ConnectionString(BSTR *pVal)
{
	(*pVal) = SysAllocString(m_bsConnectionString);

	return S_OK;
}

STDMETHODIMP CODBCClient::put_ConnectionString(BSTR newVal)
{
	if (m_bsConnectionString) SysFreeString(m_bsConnectionString);
	m_bsConnectionString = SysAllocString(newVal);

	return S_OK;
}

STDMETHODIMP CODBCClient::get_DebugLevel(short *pVal)
{
	(*pVal) = m_nDebugLevel;

	return S_OK;
}

STDMETHODIMP CODBCClient::put_DebugLevel(short newVal)
{
	m_nDebugLevel = newVal;

	return S_OK;
}

STDMETHODIMP CODBCClient::get_SocketTimeout(short *pVal)
{
	(*pVal) = m_nSocketTimeout;

	return S_OK;
}

STDMETHODIMP CODBCClient::put_SocketTimeout(short newVal)
{
	m_nSocketTimeout = newVal;

	return S_OK;
}
