/***************************************************************************
                          codbcclient.cpp  -  description
                             -------------------
    begin                : Sat Oct 30 1999
    copyright            : (C) 1999 by Team FXML
    email                : fxml@exite.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "codbcclient.h"
#include <arpa/inet.h>
#include <netinet/in.h>
#include <stdio.h>
#include <string.h>
#include <sys/socket.h>
#include <unistd.h>
#include <netdb.h>
#include <sys/time.h>
#include <sys/types.h>

CODBCClient::CODBCClient(){
	nPort = 9628;
	nSocketTimeout = 30;
}
CODBCClient::~CODBCClient(){
}

/////////////////////////////////////////////////////////////////
//
//	Name: ExecSQL
//	Params: sSQL - SQL to execute
//  Returns: XML result string
//	Description: Connects to remote host, executes SQL, returns result
/////////////////////////////////////////////////////////////////

std::string CODBCClient::ExecSQL(std::string sSQL)
{
  struct sockaddr_in address;  //address to connect to
  int sock;		//socket
	std::string sTemp;  //temp string for later manipulation
	struct in_addr inaddr; //address
	struct hostent * host; //host entry
	std::string sXML; //xml string to send
	std::string sRet = "<?xml version=\"1.0\">\r\n"; //result string
	short bIsIPAddress = 1; //if IP address or not
	short bAllDone = 0; //whether we are all done
  char *buf; //read buffer
  int amount; //amount of chars read back
  int nTemp; //temp comparison int
 	
	/* convert argument to IP or lookup DNS */
	if (!inet_aton(sHostName.c_str(), &inaddr))
	{
		//lookup host name
		host = gethostbyname(sHostName.c_str());
  	if (!host)
		{
			herror("Error looking up host name");
			sRet += "<result status=\"failure\">\r\n<error>Error looking up host</error>\r\n</result>";
		  return sRet;
		}
		bIsIPAddress = 0;
	}
	
	//create socket
	if ((sock = socket(PF_INET, SOCK_STREAM, 0)) < 0)
	{
		sRet += "<result status=\"failure\">\r\n<error>Error creating socket</error>\r\n</result>";
	  return sRet;
	}

	address.sin_family = AF_INET;
	address.sin_port = htons(nPort);

	if (!bIsIPAddress)
	{
		/*grap first IP address */
		memcpy(&address.sin_addr, host->h_addr_list[0], sizeof(address.sin_addr));
	}
	else
		address.sin_addr.s_addr = (u_long)inet_addr(sHostName.c_str());
		
	if (connect(sock, (struct sockaddr *)&address, sizeof(address)))
	{
		sRet += "<result status=\"failure\">\r\n<error>Error connecting to remote host</error>\r\n</result>";
	  return sRet;
	}
	
	//sanity check
	if (sSQL.length() == 0)
	{
		sRet += "<result status=\"failure\">\r\n<error>No SQL provided</error>\r\n</result>";
	  return sRet;
	}

	sSQL = SubStrReplace(sSQL, "&", "&amp;");
	sSQL = SubStrReplace(sSQL, "<", "&lt;");
	sSQL = SubStrReplace(sSQL, ">", "&gt;");

		
	//create the XML string to send	
	sXML = (std::string)"<?xml version=\"1.0\">\r\n<request>\r\n<connectionstring>"+ sConnectionString +
		(std::string)"</connectionstring>\r\n<sql>" + sSQL + (std::string)"</sql>\r\n</request>\r\n";
		
	//send the XML	
	write(sock, sXML.c_str(), sXML.length());
	
	//now read the response back
	fd_set read_fds;
	struct timeval TimeOut;
	FD_ZERO(&read_fds);
	FD_SET(sock, &read_fds);
	TimeOut.tv_sec = nSocketTimeout; //use timeout of durartion m_nSocketTimeout
	TimeOut.tv_usec = 0;
	sRet = "";
 	do
 	{
	 	if (select(FD_SETSIZE, &read_fds, NULL, NULL, &TimeOut) != -1)
		{
			if (FD_ISSET(sock, &read_fds))
			{
			  //copy over buffer
				buf = (char*)malloc(1000);
				amount = read(sock, buf, 1000);
 				sTemp = buf;
				sRet += sTemp.substr(0,amount);
				free(buf);
	    	if (amount < 0)
 				{
					sRet = "<?xml version=\"1.0\">\r\n<result status=\"failure\">\r\n<error>Error reading from socket</error>\r\n</result>";
					break;
				}
				//if we find result string, stop now
				nTemp = sRet.find("</result>");
				if (nTemp > 0)
				{
					bAllDone = 1;
				}
			}//if fd_isset
			else
			{
				sRet = "<?xml version=\"1.0\">\r\n<result status=\"failure\">\r\n<error>Timeout waiting for socket</error>\r\n</result>";
				break;
			}
		}//select
		else
		{
			sRet = "<?xml version=\"1.0\">\r\n<result status=\"failure\">\r\n<error>Timeout waiting for socket</error>\r\n</result>";
			break;
		}
  } while ((amount > 0) && (!bAllDone));

  shutdown(sock, 1);

	close(sock);
	if (sRet.empty()) sRet = "<?xml version=\"1.0\"?>\r\n<result status=\"failure\">\r\n<error>Communication error, connection possibly denied</error>\r\n</result>";
	return sRet;
}


///////////////////////////////////////////////////////
//
//	Name: SubStrReplace
//	Params: cOriginal - original string to search
//			cSearch - char to search for
//			cReplace - char to replace
//	Returns: string with cSearch replaced by cReplace
//	Description: Searches and replaces occurences of cSearch
//		in cOriginal with cReplace
///////////////////////////////////////////////////////
std::string CODBCClient::SubStrReplace(std::string cOriginal, char* cSearch, char* cReplace)
{
	std::string sRet = cOriginal; //set return string
	int nFind; //location of found string

	//search for char string
	nFind = sRet.find(cSearch, 0);
	while (nFind >= 0)
	{
		sRet = sRet.substr(0, nFind) + (std::string)cReplace + sRet.substr(nFind + strlen(cSearch), sRet.length() - nFind);

		nFind = sRet.find(cSearch, nFind + strlen(cReplace));
	}

	return sRet;
}


